"""Property prediction algorithms."""

from typing import TypeVar

# from ..core import PropertyPredictor
# from ...domains.materials import ConditionPAG


S = TypeVar("S")
# T = TypeVar('T')  CLaSSVAE uses List[PropertyPredictor[S, U]] as target
U = TypeVar("U")
