"""Not so domain specific code."""

# NOTE these might change a lot
# - could be abstract base classes such that other implmentations
#   can be simply registered, without explicit inheritance
# - could be simply NewType definitions if they do not implement functionality
# - organize/sort into domain specific files
